#!BPY
# vim:encoding=ISO-8859-1

"""
Widgets for ScriptUI.

@version: 0.9.2
@author:  Mathias Panzenbck
@license: LGPL

@group Base Widgets: Widget, Rect, SimpleWidget, PushButton, Number,
	Slider, String,	Menu, Toggle, Image, SimpleText

@group Advanced Widgets: PixelField, RenderField, ScrollText, FileField

@group Container Widgets: ContainerWidget, RootWidget, WidgetFrame,
	WidgetList, CollapseWidget

@group Helper Functions: split_to_fit, next_word_end
"""

__author__  = "Mathias Panzenbck"
__version__ = "0.9.2"
__url__     = ("blender","elysiun","http://www.elysiun.com/forum/viewtopic.php?t=38458&postdays=0&postorder=asc&start=0",)
__license__ = "LGPL"
__bpydoc__  = """\
Description: Wrapper to blenders script-UI-eventsystem
Usage: only used by other scripts
"""

# --------------------------------------------------------------------------
# ScriptUI.Widgets, Copyright (C) 2005  Mathias Panzenbck
# --------------------------------------------------------------------------
# ***** BEGIN LGPL LICENSE BLOCK *****
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# ***** END LGPL LICENCE BLOCK *****
# -------------------------------------------------------------------------- 

import Blender
from   Blender.BGL    import *
from   Blender.Draw   import *
from   Blender.Window import *

import ScriptUI
from ScriptUI import *

import ScriptUI.Color
from ScriptUI.Color import *

class Widget:
	"""
	Base class for all widgets.
	"""
	
	def __init__(
			self,
			x = 0,
			y = 0,
			width   = 50,
			height  = 50,
			parent  = None,
			evtmngr = None):
		"""
		Initialise a Widget.
		
		When parent != None and self != self.parent, the widget will add it self to
		it's parent.
		
		When evtmngr == None and parent != None, the widget will use the
		L{ScriptUI.EventManager} of it's parent.
		
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		if parent != None:
			self.parent = parent

			if not evtmngr:
				evtmngr = parent.evtmngr
		else:
			if not evtmngr:
				self.parent = None
			else:
				self.parent = evtmngr.noroot

# don't do that, so you can create widgets, which 
# uses the right EventManager but aren't childs
# of any ContainerWidget.
# this is usefull when you pre-create widgets in
# the background and only add them to a container
# when they are needed.
# hmm, will such widgets be really deleted on exit?
# even if they have childs (backref)?
# have to FIX this..
#
#		else:
#			self.parent = evtmngr.root
		
		self.evtmngr = evtmngr

		if self != self.parent and self.parent != None:
			self.parent.add( self )
		
		self.x       = int(x)
		self.y       = int(y)
		self._width  = int(width)
		self._height = int(height)
		self.visible = True
	
	def __del__(self):
		"""
		Destructor.

		Calls self.unref().
		"""
		self.unref()
	
	def up(self):
		if self.parent:
			self.parent.move_up(self)
			
	def down(self):
		if self.parent:
			self.parent.move_down(self)
			
	def top(self):
		if self.parent:
			self.parent.move_top(self)
			
	def bottom(self):
		if self.parent:
			self.parent.move_bottom(self)

	def draw(self,x,y):
		"""
		The draw function will be called on every redraw.

		If you derive your own widget and need to do (e.g. draw) something
		on every redraw, overload this function. Be sure to call the base-classes
		draw function, too. e.g.: Widget.draw(self,x,y)

		@type  x: int
		@param x: The absolut x position of the bottom-left-corner of your widget.
		@type  y: int
		@param y: The absolut y position of the bottom-left-corner of your widget.
		"""
		pass

	def unref(self):
		"""
		Unref all child-widgets.

		The python garbadge collector works by ref-count.
		This unref method has to unref (set to None) all child-widgets
		and also it's parent, so no memleaks would happen.

		When you derive your own widget and you store references to child-widgets,
		be sure to overload this function and set the references to None. Don't
		forget to call the unref function of the base-class. e.g.: Widget.unref(self)
		"""
		self.parent  = None
		self.evtmngr = None

	def get_width(self):
		"""
		Get the width of the widget.

		Do not access the _width attribute directly, because some widgets don't depend
		on it and calculate the width dynamically.

		@rtype:  int
		@return: The widgets width.
		"""
		return self._width
	
	def set_width(self,val):
		"""
		Set the width of the widget.

		Do not access the _width attribute directly, because some widgets don't depend
		on it and calculate the width dynamically (and ignore the call of set_width).

		@type val: int
		"""
		mc = self.evtmngr.MakeChanges()
		
		self._width = int(val)
	
	def get_height(self):
		"""
		Get the height of the widget.

		Do not access the _height attribute directly, because some widgets don't depend
		on it and calculate the height dynamically.
		
		@rtype:  int
		@return: The widgets height.
		"""
		return self._height

	def set_height(self,val):
		"""
		Set the height of the widget.

		Do not access the _height attribute directly, because some widgets don't depend
		on it and calculate the height dynamically (and ignore the call of set_height).

		@type val: int
		"""
		mc = self.evtmngr.MakeChanges()
		
		self._height = int(val)

	def get_rect(self):
		"""
		Returns the bounding-rectangle of the widget.

		@rtype:  (int,int,int,int)
		@return: (x1,y1,x2,y2)
		"""
		return (self.x, self.y, self.get_width() + self.x, self.get_height() + self.y)

	def set_rect(self,(x1,y1,x2,y2)):
		"""
		Sets x, y, width and height according the given bounding-rectangel.

		@type x1: int
		@type y1: int
		@type x2: int
		@type y2: int
		"""
		mc = self.evtmngr.MakeChanges()
		
		if x1 > x2: x1, x2 = x2, x1
		if y1 > y2: y1, y2 = y2, y1
			
		self.x = int(x1)
		self.y = int(y1)
		self.set_width(x2 - x1)
		self.set_height(y2 - y1)
	
	def get_visible(self):
		"""
		Get the visibility state of the widget.

		@rtype: bool
		"""
		return self.visible

	def set_visible(self,val=True):
		"""
		Set the visibility state of the widget.

		@type val: bool
		"""
		mc = self.evtmngr.MakeChanges()

		self.visible = bool(val)

	def get_parent(self):
		"""
		Get widget's parent.

		@rtype: Widget
		"""
		return self.parent

	def set_parent(self,parent):
		"""
		Set widgets parent.

		@type parent: Widget
		"""
		mc = self.evtmngr.MakeChanges()
		
		if self.parent:
			self.parent.remove(self)

		if parent == None:
			self.parent = None
		else:
			parent.add(self)

class ContainerWidget(Widget):
	"""
	Base-class for all Wdgets which do contain child-widgets.

	All container widgets emulate sequences. At least the __len__ and __iter__
	methods are overloaded, so you can access all childs be using a for-loop.
	"""
	def __init__(
			self,
			childs = [],
			x = 0,
			y = 0,
			width   = 100,
			height  = 100,
			parent  = None,
			evtmngr = None):
		"""
		@type  childs: list
		@param childs: list of child-widgets.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		self._childs = []

		Widget.__init__(self,x,y,width,height,parent,evtmngr)

		self.add( childs )
	
	def unref(self):
		for child in self:
			child.unref()

		self._childs = []
		
		Widget.unref(self)
		
	def index(self,child):
		return self._childs.index(child)

	def move_up(self,child):
		ind = self.index(child)

		if ind + 1 < len(self):
			wdg1 = self._childs[ind]
			wdg2 = self._childs[ind+1]

			self._childs[ind+1] = wdg1
			self._childs[ind]   = wdg2
	
	def move_down(self,child):
		ind = self.index(child)

		if ind > 0:
			wdg1 = self._childs[ind]
			wdg2 = self._childs[ind-1]

			self._childs[ind-1] = wdg1
			self._childs[ind]   = wdg2
			
	def move_top(self,child):
		ind = self.index(child)
		top = len(self) - 1
		
		wdg1 = self._childs[ind]
		wdg2 = self._childs[top]
		
		self._childs[top] = wdg1
		self._childs[ind] = wdg2
			
	def move_bottom(self,child):
		ind = self.index(child)
		
		wdg1 = self._childs[ind]
		wdg2 = self._childs[0]
		
		self._childs[0]   = wdg1
		self._childs[ind] = wdg2
	
	def clear(self):
		"""
		Clears all child-widgets.
		"""
		mc = self.evtmngr.MakeChanges()

		for child in self:
			child.parent = None

		self._childs = []

	def draw(self,x,y):
		if self.visible:
			cld_x = self.x + x
			cld_y = self.y + y

			for child in self:
				child.draw(cld_x,cld_y)

	def remove(self,widget):
		"""
		Remove a child-widget.

		Removes 'widget' from childs-list and sets widgets parent to None.

		@type  widget: Widget
		@param widget: The widget to remove.
		"""
		mc = self.evtmngr.MakeChanges()

		self._childs.remove(widget)
		widget.parent = None

	def __iter__(self):
		return iter(self._childs)

	def __len__(self):
		return len(self._childs)

	def __getitem__(self,key):
		return self._childs[key]

	def add(self,child):
		"""
		Add one or more child-widget.

		If the new child has a parent, it will be removed from it.

		@type  child: Widget or list
		@param child: The child(s) to add.
		"""
		mc = self.evtmngr.MakeChanges()
		
		if type(child) == list:
			for cld in child:
				if cld.parent:
					try:
						cld.parent.remove(cld)
					except ValueError:
						pass
				cld.parent = self

			self._childs += child
		else:
			if child.parent:
				try:
					child.parent.remove(child)
				except ValueError:
					pass
			child.parent = self
			self._childs.append(child)
	
	def is_empty(self):
		"""
		Returns True if there are no childs.
		
		@rtype: bool
		"""
		return self._childs == []

	def get_scale_width(self):
		"""
		Returns width of the area where childs can be placed.

		This is usefull for container-widgets which support a margin.

		@rtype: int
		"""
		return self.get_width()
	
	def get_scale_height(self):
		"""
		Returns height of the area where childs can be placed.

		This is usefull for container-widgets which support a margin.

		@rtype: int
		"""
		return self.get_height()

	def get_scale_left(self):
		"""
		Returns left most position where childs can be placed.

		This is usefull for container-widgets which support a margin.

		@rtype: int
		"""
		return 0
	
	def get_scale_bottom(self):
		"""
		Returns bottom most position where childs can be placed.

		This is usefull for container-widgets which support a margin.

		@rtype: int
		"""
		return 0

class RootWidget(ContainerWidget):
	"""
	The root-widget.

	Do not make an instance of this class, it's only used internally by the L{ScriptUI.EventManager}.
	"""
	def __init__(self,evtmngr):
		ContainerWidget.__init__(self,[],0,0,-1,-1,self,evtmngr)
	
	def __del__(self):
		ContainerWidget.__del__(self)

	def get_height(self):
		w, h = GetAreaSize()
		return h
	
	def get_width(self):
		w, h = GetAreaSize()
		return w

class SimpleWidget(Widget):
	"""
	Base-class for all simple widgets.
	"""
	def __init__(
			self,
			name,
			x = 0,
			y = 0,
			width  = 100,
			height = 100,
			action = NOOP,
			tooltip   = "",
			initvalue = None,
			parent    = None,
			evtmngr   = None):
		"""
		@type  name: string
		@param name: The name (caption) of the widget.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  action: function with no parameter
		@param action: The action-callback-function.
		@type  tooltip: string
		@param tooltip: The widgets tooltip.
		@param initvalue: The initial value of the widget.
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		self.name        = str(name)

		Widget.__init__(self,x,y,width,height,parent,evtmngr)

		self.tooltip     = str(tooltip)
		self.widget      = Create(initvalue)
		self.action      = action
		self.event_code  = self.evtmngr.register_button_event(self._bevent)
	
	def unref(self):
		if self.event_code > -1:
			self.evtmngr.unregister_button_event(self.event_code)
			self.event_code = -1
			
		self.action = None

		Widget.unref(self)

	def _bevent(self,evt):
		self.action()
	
	def get_value(self):
		"""
		Get the widgets value.

		@return: The widgets value.
		"""
		return self.widget.val
	
	def set_value(self,val):
		"""
		Set the widgets value.

		@param val: The new value.
		"""
		mc = self.evtmngr.MakeChanges()
		
		self.widget.val = val
		self.action()
	value = property(get_value, set_value)

class PushButton(SimpleWidget):
	"""
	Create a push-button.
	"""
	def __init__(
			self,
			name,
			x = 0,
			y = 0,
			width   = 100,
			height  = 100,
			action  = NOOP,
			tooltip = "",
			parent  = None,
			evtmngr = None):
		"""
		@type  name: string
		@param name: The name (caption) of the widget.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  action: function with no parameter
		@param action: The action-callback-function.
		@type  tooltip: string
		@param tooltip: The widgets tooltip.
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		SimpleWidget.__init__(self,name,x,y,width,height,action,tooltip,0,parent,evtmngr)

	def draw(self,x,y):
		if self.visible:
			self.widget = Blender.Draw.PushButton(
				self.name,
				self.event_code,
				self.x + x,
				self.y + y,
				self._width,
				self._height,
				self.tooltip)
			
	def get_value(self):
		return None
	
	def set_value(self,val):
		pass

class Number(SimpleWidget):
	"""
	Create a number-widget for int or float input.
	"""
	def __init__(self, name,
			x = 0, y = 0,
			width = 100, height = 100,
			action = NOOP, initvalue = 1,
			min = 0, max = 1,
			tooltip = "",
			parent  = None, evtmngr = None):
		"""
		@type  name: string
		@param name: The name (caption) of the widget.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  action: function with no parameter
		@param action: The action-callback-function.
		@type  initvalue: int or float
		@param initvalue: The initial value of the widget.
		@type  min: int or float
		@param min: Minimum value.
		@type  max: int or float
		@param max: Maximum value.
		@type  tooltip: string
		@param tooltip: The widgets tooltip.
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		SimpleWidget.__init__(self,name,x,y,width,height,action,tooltip,initvalue,parent,evtmngr)

		self.min = float(min)
		self.max = float(max)
	
	def draw(self,x,y):
		if self.visible:
			self.widget = Blender.Draw.Number(
				self.name,
				self.event_code,
				self.x + x,
				self.y + y,
				self._width,
				self._height,
				self.widget.val,
				self.min,
				self.max,
				self.tooltip)

class Slider(SimpleWidget):
	"""
	Create a slider-widget for int or float input.
	"""
	def __init__(self, name,
			x = 0, y = 0,
			width = 100, height = 100,
			action = NOOP, initvalue = 1.,
			min = 0., max = 1., realtime = 0,
			tooltip = "",
			parent  = None, evtmngr = None):
		"""
		@type  name: string
		@param name: The name (caption) of the widget.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  action: function with no parameter
		@param action: The action-callback-function.
		@type  initvalue: int or float
		@param initvalue: The initial value of the widget.
		@type  min: int or float
		@param min: Minimum value.
		@type  max: int or float
		@param max: Maximum value.
		@type  realtime: int
		@param realtime: If non-zero, the slider will emit events as it is edited.
		@type  tooltip: string
		@param tooltip: The widgets tooltip.
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		SimpleWidget.__init__(self,name,x,y,width,height,action,tooltip,initvalue,parent,evtmngr)

		if initvalue > max:
			max = initvalue
		self.min      = min
		self.max      = max
		self.realtime = int(realtime)
	
	def draw(self,x,y):
		if self.visible:
			self.widget = Blender.Draw.Slider(
				self.name,
				self.event_code,
				self.x + x,
				self.y + y,
				self._width,
				self._height,
				self.widget.val,
				self.min,
				self.max,
				self.realtime,
				self.tooltip)

class String(SimpleWidget):
	"""
	Create a string-input widget.
	"""
	def __init__(self, name,
			x = 0, y = 0,
			width = 100, height = 100,
			action = NOOP, initvalue = "",
			length = 512,
			tooltip = "",
			parent  = None, evtmngr = None):
		"""
		@type  name: string
		@param name: The name (caption) of the widget.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  action: function with no parameter
		@param action: The action-callback-function.
		@type  initvalue: string
		@param initvalue: The initial value of the widget.
		@type  length: int
		@param length: The maximal length of the input-string.
		@type  tooltip: string
		@param tooltip: The widgets tooltip.
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		SimpleWidget.__init__(self,name,x,y,width,height,action,tooltip,initvalue,parent,evtmngr)

		self.length = int(length)
	
	def draw(self,x,y):
		if self.visible:
			self.widget = Blender.Draw.String(
				self.name,
				self.event_code,
				self.x + x,
				self.y + y,
				self._width,
				self._height,
				self.widget.val,
				self.length,
				self.tooltip)
			
	def set_value(self,val):
		SimpleWidget.set_value(self,str(val))

class Toggle(SimpleWidget):
	"""
	Create a toggle-button widget.
	"""
	def __init__(self, name,
			x = 0, y = 0,
			width = 100, height = 100,
			action = NOOP, initvalue = 0,
			tooltip = "",
			parent  = None, evtmngr = None):
		"""
		@type  name: string
		@param name: The name (caption) of the widget.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  action: function with no parameter
		@param action: The action-callback-function.
		@type  initvalue: int
		@param initvalue: The initial value of the widget.
		@type  tooltip: string
		@param tooltip: The widgets tooltip.
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		SimpleWidget.__init__(self,name,x,y,width,height,action,tooltip,int(initvalue),parent,evtmngr)
	
	def draw(self,x,y):
		if self.visible:
			self.widget = Blender.Draw.Toggle(
				self.name,
				self.event_code,
				self.x + x,
				self.y + y,
				self._width,
				self._height,
				self.widget.val,
				self.tooltip)
	
	def set_value(self,val):
		SimpleWidget.set_value(self,int(val))
	
class Menu(Widget):
	"""
	Create a Menu-widget.
	
	@cvar ITEM: Defines a menu-entry as "item".
	@type ITEM: int
	@cvar SEP:  Defines a menu-entry as "seperator".
	@type SEP:  int
	"""
	
	ITEM = MenuBase.ITEM
	SEP  = MenuBase.SEP
	
	def __init__(
			self,
			name = None,
			x = 0,
			y = 0,
			width   = 100,
			height  = 100,
			items   = [],
			initvalue = 0,
			tooltip = "",
			parent  = None,
			evtmngr = None):
		"""
		@type  name: string
		@param name: The name (caption) of the widget.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  items: list
		@param items: Menu-items (see: L{ScriptUI.MenuBase})
		@type  initvalue: int
		@param initvalue: The initial selected menu-entry.
		@type  tooltip: string
		@param tooltip: The widgets tooltip.
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		Widget.__init__(self,x,y,width,height,parent,evtmngr)

		self.widget     = Create(initvalue)
		self.tooltip    = tooltip
		self.event_code = self.evtmngr.register_button_event(self._bevent)
		self.menu       = MenuBase(name,items)

	def unref(self):
		if self.event_code > -1:
			self.menu.clear()
			self.menu = None
			self.evtmngr.unregister_button_event(self.event_code)
			self.event_code = -1

		Widget.unref(self)

	def get_name(self):
		"""
		Returns the name (title) of the menu.

		@rtype:  string
		@return: The name (title) of the menu.
		"""
		return self.menu.name
	
	def set_name(self,name):
		"""
		Sets the name (title) of the menu.

		@type  name: string
		@param name: The name (title) of the menu.
		"""
		self.menu.name = str(name)

	def get_value(self):
		"""
		Returns the index of the selected menu-entry.

		@rtype:  int
		@return: The index of the selected menu-entry.
		"""
		return self.widget.val
	
	def set_value(self,val):
		"""
		Sets the index of the selected menu-entry.

		@type  val: int
		@param val: The index of the selected menu-entry.
		"""
		mc = self.evtmngr.MakeChanges()
		
		self.widget.val = int(val)
		self.menu(val)
	value = property(get_value, set_value)

	def addItem(self,name,clbk=NOOP):
		"""
		Add a new menu-entry.

		@type  name: string
		@param name: The name (title) of the new menu-entry.
		@type  clbk: function with no parameters
		@param clbk: The function to call when the menu-entry was clicked.
		"""
		mc = self.evtmngr.MakeChanges()
		
		self.menu.addItem(name,clbk)

	def addSep(self):
		"""
		Add a new seperator.
		"""
		mc = self.evtmngr.MakeChanges()
		
		self.menu.addSep()
		
	def _bevent(self,evt):
		self.menu( self.widget.val )
		
	def draw(self,x,y):
		if self.visible:
			self.widget = Blender.Draw.Menu(
				self.menu.get_menu_str(),
				self.event_code,
				self.x + x,
				self.y + y,
				self._width,
				self._height,
				self.widget.val,
				self.tooltip)

class Image(Widget):
	"""
	Draw an image on the screen.

	The image is drawn at the location specified by the coordinates (x,y). A pair of optional zoom
	factors (in horizontal and vertical directions) can be applied to the image as it is drawn, and
	an additional clipping rectangle can be applied to extract a particular sub-region of the image
	to draw.

	Note that the clipping rectangle is given in image space coordinates. In image space, the origin
	is located at the bottom left, with x coordinates increasing to the right and y coordinates
	increasing upwards. No matter where the clipping rectangle is placed in image space, the
	lower-left pixel drawn on the screen is always placed at the coordinates (x,y). The clipping
	rectangle is itself clipped to the dimensions of the image. If either the width or the height
	of the clipping rectangle are negative then the corresponding dimension (width or height) is set
	to include as much of the image as possible.

	@todo: test Image widget.
	"""
	def __init__(self, image,
			x = 0, y = 0,
			zoomx = 1.0, zoomy = 1.0,
			clipx = 0, clipy = 0,
			clipw = 100, cliph = 100,
			parent = None, evtmngr = None):
		"""
		@param image: The image to draw.
		@type  image: Blender.Image 
		@param x: The lower left x (horizontal) position of the origin of the image.
		@type  x: int 
		@param y: The lower left y (vertical) position of the origin of the image.
		@type  y: int 
		@param zoomx: The x (horizontal) zoom factor to use when drawing the image.
		@type  zoomx: float 
		@param zoomy: The y (vertical) zoom factor to use when drawing the image.
		@type  zoomy: float 
		@param clipx: The lower left x (horizontal) origin of the clipping rectangle within the image.
			A value of 0 indicates the left of the image.
		@type  clipx: int 
		@param clipy: The lower left y (vertical) origin of the clipping rectangle within the image.
			A value of 0 indicates the bottom of the image.
		@type  clipy: int 
		@param clipw: The width of the clipping rectangle within the image.
			If this value is negative then the clipping rectangle includes
			as much of the image as possible in the x (horizontal) direction.
		@type  clipw: int 
		@param cliph: The height of the clipping rectangle within the image.
			If this value is negative then the clipping rectangle includes
			as much of the image as possible in the y (vertical) direction.
		@type  cliph: int
		@param parent: The parent widget.
		@type  parent: Widget
		@param evtmngr: The EventManager to use.
		@type  evtmngr: ScriptUI.EventManager
		"""
		Widget.__init__(self,x,y,clipw,cliph,parent,evtmngr)

		self.clipx  = int(clipx)
		self.clipy  = int(clipy)
		self.zoomx  = float(zoomx)
		self.zoomy  = float(zoomy)
		self.image  = image
		self.widget = Create(0)
	
	def draw(self,x,y):
		if self.visible:
			self.widget = Blender.Draw.Image(self.image,
					self.x + x, self.y + y,
					self.zoomx, self.zoomy,
					self.clipx, self.clipy,
					self._width, self._height)

class PixelField(Widget):
	"""
	Draw a field of RGBA-pixels to the screen.

	@ivar pixels: RGBA-pixel array.
	@type pixels: Blender.BGL.Buffer(Blender.BGL.GL_FLOAT, [height, width, 4])
	"""
	def __init__(self,
			x = 0, y = 0,
			width = 100, height = 100,
			parent = None, evtmngr = None):
		"""
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		Widget.__init__(self,x,y,width,height,parent,evtmngr)
		
		self.pixels = Blender.BGL.Buffer(Blender.BGL.GL_FLOAT, [height, width, 4])

	def draw(self,x,y):
		if self.visible:
			rndr_x = self.x + x
			rndr_y = self.y + y
			
			Blender.BGL.glRasterPos2i(rndr_x, rndr_y)
			Blender.BGL.glDrawPixels(self.get_width(), self.get_height(), Blender.BGL.GL_RGBA, Blender.BGL.GL_FLOAT, self.pixels)

class RenderField(Widget):
	"""
	Render a function to screen.

	Like L{PixelField} but on every redraw the render-callback-function will be called
	for each pixel.
	"""
	def __init__(self,render_clbk,
			x = 0, y = 0,
			width = 100, height = 100,
			parent = None, evtmngr = None):
		"""
		@type  render_clbk: function with two parametrs: x (int), y (int)
		@param render_clbk: The render-callback-function.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		Widget.__init__(self,x,y,width,height,parent,evtmngr)

		self.render_clbk = render_clbk

	def draw(self,x,y):
		if self.visible:
			# render_clbk(x,y):
			# x, y are in range [0,1]

			w      = self.get_width()
			h      = self.get_height()
			dx     = 1. / w
			dy     = 1. / h
			rndr_x = self.x + x
			rndr_y = self.y + y
			
			pixels = Blender.BGL.Buffer(Blender.BGL.GL_FLOAT, [h, w, 1])

			px = 0.
			for i in xrange(w):
				py = 0.
				for j in xrange(h):
					pixels[j][i][0] = self.render_clbk(px,py)
					py += dy
				px += dx

			Blender.BGL.glRasterPos2i(rndr_x, rndr_y)
			Blender.BGL.glDrawPixels(w, h, Blender.BGL.GL_LUMINANCE, Blender.BGL.GL_FLOAT, pixels)

class SimpleText(Widget):
	"""
	Print text to screen.

	Use this to print a simple, multiline text to screen.
	width and height are not setable, they will be calculated from the lines of the text.
	"""
	def __init__(
			self,
			text  = "",
			style = "normal",
			line_height = 15,
			bottom_up   = False,
			color       = WHITE,
			x = 0,
			y = 0,
			parent  = None,
			evtmngr = None):
		"""
		@type  text: string or list
		@param text: The text to draw.
		@type  style: string
		@param style: The size of the font: 'large', 'normal', 'small' or 'tiny'. 
		@type  line_height: int
		@param line_height: The height of a line.
		@type  bottom_up: bool
		@param bottom_up: True to start with first line at the bottom.
		@type  color: ScriptUI.Color.Color
		@param color: The color of the text.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		Widget.__init__(self,x,y,0,0,parent,evtmngr)

		if type(text) == str:
			text = text.split("\n")

		self.lines       = [line.rstrip() for line in text]
		self.style       = str(style)
		self.bottom_up   = bool(bottom_up)
		self.line_height = int(line_height)
		self.color       = Color(*color)
	
	def draw(self,x,y):
		if not self.visible:
			return

		h        = self.get_height()
		oldcolor = getCurrentColor()
		pos      = self.y + y - self.line_height
		x        = self.x + x

		if self.bottom_up:
			pos += h

		glColor4f( *self.color )
		
		for line in self.lines:
			glRasterPos2i( x, pos )
			Text( line, self.style )
			pos -= self.line_height

		glColor4f( *oldcolor )
	
	def set_text(self,text):
		"""
		Set a new text.

		@type  text: string or list
		@param text: The new text.
		"""
		mc = self.evtmngr.MakeChanges()
		
		if type(text) == str:
			text = text.split("\n")
		
		self.lines = [line.rstrip() for line in text]
	
	def set_height(self,val):
		pass
	
	def set_width(self,val):
		pass

	def get_width(self):
		mw = 0
		for line in self.lines:
			w = Blender.Draw.GetStringWidth(line)
			
			if mw < w:
				mw = w
		return w
	
	def get_height(self):
		return self.line_height * len(self.lines)

class ScrollText(ContainerWidget):
	"""
	Create a scrolleable text-field.

	@cvar BUTTON_WIDTH:  Width of the button 'UP' and 'DOWN'.
	@cvar BUTTON_HEIGHT: Height of the button 'UP' and 'DOWN'.
	"""
	BUTTON_WIDTH  = 55
	BUTTON_HEIGHT = 18
		
	def __init__(
			self,
			text  = "",
			style = "normal",
			line_height = 15,
			margin  = 5,
			color   = BLACK,
			bgcolor = GREY,
			x = 0,
			y = 0,
			width   = 100,
			height  = 36,
			parent  = None,
			evtmngr = None):
		"""
		@type  text: string
		@param text: The text to draw.
		@type  style: string
		@param style: The size of the font: 'large', 'normal', 'small' or 'tiny'. 
		@type  line_height: int
		@param line_height: The height of a line.
		@type  margin: int
		@param margin: Margin around the text.
		@type  color: ScriptUI.Color.Color
		@param color: The color of the text.
		@type  bgcolor: ScriptUI.Color.Color
		@param bgcolor: The background-color.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		if type(text) == str:
			text = text.split("\n")

		self.lines       = [line.rstrip() for line in text]
		self.line_height = line_height
		self.style       = style
		self.margin      = margin
		self.bgcolor     = Color(*bgcolor)
		self.color       = Color(*color)

		self.scroll_pos = 0
		self.scroll_max = len(self.lines) - 1
		
		ContainerWidget.__init__(self,[],x,y,width,height,parent,evtmngr)

		self._btn_up = PushButton(
				"UP",
				width - ScrollText.BUTTON_WIDTH,
				height - ScrollText.BUTTON_HEIGHT,
				ScrollText.BUTTON_WIDTH,
				ScrollText.BUTTON_HEIGHT,
				self.scroll_up,
				tooltip = "scroll up",
				parent  = self,
				evtmngr = evtmngr)

		self._btn_down = PushButton(
				"DOWN",
				width - ScrollText.BUTTON_WIDTH,
				0,
				ScrollText.BUTTON_WIDTH,
				ScrollText.BUTTON_HEIGHT,
				self.scroll_down,
				tooltip = "scroll down",
				parent  = self,
				evtmngr = evtmngr)

	def unref(self):
		if self._btn_up:   self._btn_up.unref()
		if self._btn_down: self._btn_down.unref()
		
		self._btn_up   = None
		self._btn_down = None

		ContainerWidget.unref(self)
		
	def clear(self):
		"""
		Clears the text and the scroll-position.
		"""
		mc = self.evtmngr.MakeChanges()
		
		self.scroll_pos =  0
		self.scroll_max = -1
		self.lines      = []
		
	def set_width(self,val):
		mc = self.evtmngr.MakeChanges()
		
		Widget.set_width(self,val)

		x = int(val) - ScrollText.BUTTON_WIDTH
		self._btn_up.x   = x
		self._btn_down.x = x
	
	def set_height(self,val):
		mc = self.evtmngr.MakeChanges()
		
		Widget.set_height(self,val)

		self._btn_up.y = int(val) - ScrollText.BUTTON_HEIGHT

	def scroll_top(self):
		"""
		Scrolls to the top.
		"""
		if self.scroll_pos != 0:
			mc = self.evtmngr.MakeChanges()
			self.scroll_pos = 0
		
	def scroll_bottom(self):
		"""
		Scrolls to the bottom.
		"""
		if self.scroll_pos != self.scroll_max:
			mc = self.evtmngr.MakeChanges()
			self.scroll_pos = self.scroll_max

	def scroll_up(self):
		"""
		Scrolls one line up.
		"""
		if self.scroll_pos > 0:
			mc = self.evtmngr.MakeChanges()
			self.scroll_pos -= 1

	def scroll_down(self):
		"""
		Scrolls one line down.
		"""
		if self.scroll_pos < self.scroll_max:
			mc = self.evtmngr.MakeChanges()
			self.scroll_pos += 1

	def set_text(self,text):
		"""
		Set a new text.

		@type  text: string or list
		@param text: The new text.
		"""
		mc = self.evtmngr.MakeChanges()
		
		if type(text) == str:
			text = text.split("\n")
		
		self.lines = [line.rstrip() for line in text]
			
		self.scroll_pos = 0
		self.scroll_max = len(self.lines) - 1
		
	def draw(self,glob_x,glob_y):
		if not self.visible:
			return
		
		x          = self.x + glob_x
		y          = self.y + glob_y
		text_x     = x + self.margin
		draw_width = self._width - 2 * self.margin
		# text geht ja auch nach oben, wie alles in blender.
		pos        = y + self._height - self.margin - self.line_height - ScrollText.BUTTON_HEIGHT
		maxpos     = y + self.margin + ScrollText.BUTTON_HEIGHT
		oldcolor   = getCurrentColor()

		glColor4f( *self.bgcolor )
		glRecti( x, y, x + self._width, y + self._height )
		glColor4f( *self.color )
				
		linenum = 0
		cancel  = False
		for line in self.lines:
			if linenum >= self.scroll_pos:
				text = split_to_fit(line,draw_width,self.style)

				for broken_line in text:
					if pos <= maxpos:
						cancel = True
						break
					glRasterPos2i( text_x, pos )
					Text( broken_line, self.style )
					pos -= self.line_height
			linenum += 1

			if cancel:
				break
			
		glColor4f( *oldcolor )
		
		ContainerWidget.draw(self,glob_x,glob_y)

	def get_scale_width(self):
		return self._width - 2 * self.margin
	
	def get_scale_height(self):
		return self._height - 2 * self.margin
	
	def get_scale_left(self):
		return self.margin
	
	def get_scale_bottom(self):
		return self.margin

class FileField(ContainerWidget):
	"""
	Create a file-chooser field.

	@ivar selector: Filedialog to use.
	@cvar BUTTON_WIDTH: Width of the 'browse...' button.
	@cvar FILE:  If L{selector} == FILE, use FileSelector.
	@cvar IMAGE: If L{selector} == IMAGE, use ImageSelector.
	"""
	BUTTON_WIDTH = 70

	FILE  = Blender.Window.FileSelector
	IMAGE = Blender.Window.ImageSelector
	
	def __init__(
			self,
			caption = "File: ",
			browse_caption = "Open File",
			path    = "",
			x = 0,
			y = 0,
			width    = 200,
			height   =  20,
			action   = NOOP,
			selector = None,
			tooltip  = "",
			parent   = None,
			evtmngr  = None):
		"""
		@type  caption: string
		@param caption: The caption of the String-widget where you can enter the filename.
		@type  browse_caption: string
		@param browse_caption: Text to show on the 'Open/Save File' button in the filebrowser.
		@type  path: string
		@param path: Filepath.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  action: function with one parameter: filename (string)
		@param action: This function will be called when the filepath is changed
			(by the String-widget or filebrowser).
		@type  selector: function with 3 parameters: callback, title (string), path (string)
		@param selector: Selector function to use. Either L{FileField.FILE} or L{FileField.IMAGE}.
		@type  tooltip: string
		@param tooltip: The tooltip of the String-widget.
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		ContainerWidget.__init__(self,[],x,y,width,height,parent,evtmngr)
		
		if selector == None:
			self.selector = FileField.FILE
		else:
			self.selector = selector
			
		self.action         = action
		self.browse_caption = str(browse_caption)

		self._path = String(
				caption,
				0, 0,
				width - FileField.BUTTON_WIDTH,	height,
				self._bevent,	path, 512, tooltip,
				self, evtmngr )
		
		self._btn  = PushButton(
				"browse...",
				width - FileField.BUTTON_WIDTH, 0,
				FileField.BUTTON_WIDTH, height,
				self._browse, "open file-browser",
				self, evtmngr )
		
	def unref(self):
		self.action   = None
		self.selector = None
		self._path    = None
		self._btn     = None

		ContainerWidget.unref(self)
		
	def set_width(self,val):
		mc = self.evtmngr.MakeChanges()
		
		val = int(val)
		Widget.set_width(self,val)

		self._path.set_width( val - FileField.BUTTON_WIDTH )
		self._btn.x = val - FileField.BUTTON_WIDTH
		
	def set_height(self,val):
		mc = self.evtmngr.MakeChanges()
		
		val = int(val)
		Widget.set_height(self,val)

		self._path.set_height(val)
		self._btn.set_height(val)
		
	def _bevent(self):
		self.action( self.get_path() )
		
	def _browse(self):
		path = self.get_path()

		if path != "":
			self.selector( self.set_path, self.browse_caption, path )
		else:
			self.selector( self.set_path, self.browse_caption )
			
	def get_path(self):
		"""
		Get the filepath.

		@rtype:  string
		@return: The filepath.
		"""
		return self._path.get_value()
	
	def set_path(self,path):
		"""
		Set the filepath.

		@type  path: string
		@param path: The filepath.
		"""
		mc = self.evtmngr.MakeChanges()
		
		# calls _bevent implicit
		self._path.set_value( path )


class CollapseWidget(ContainerWidget):
	"""
	Create a collapseable widget.
	"""
	BUTTON_SIZE = 15

	def __init__(self,
			x = 0,
			y = 0,
			width    = 100,
			height   = 100,
			margin   = 5,
			child    = None,
			collapse = NOOP,
			parent   = None,
			evtmngr  = None):
		"""
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  margin: int
		@param margin: The margin around the child-widgets.
		@type  child: Widget
		@param child: The child-widget.
		@type  collapse: function with one parameter: collapsed (bool)
		@param collapse: Will be called if collpsed-state changes.
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		ContainerWidget.__init__(self,[],x,y,width,height,parent,evtmngr)

		self.collapse = collapse
		self.margin   = int(margin)
		self._child   = child
		
		self._btn = Toggle("", 0, 0,
				CollapseWidget.BUTTON_SIZE, CollapseWidget.BUTTON_SIZE,
				self._toggle_clicked, 0, "expand/collapse widget",
				self, evtmngr)
		
		if child != None:
			self.add( child )

	def unref(self):
		self.collapse = None
		self._child   = None
		self._btn     = None

		ContainerWidget.unref(self)
	
	def _toggle_clicked(self):
		self.evtmngr.MakeChanges()

	def draw(self,x,y):
		if self.visible:
			if not self.get_collapsed() and self._child != None:
				self._child.draw(self.x + x + 2 * self.margin +
					CollapseWidget.BUTTON_SIZE,
					self.y + y + self.margin)
			self._btn.draw(self.x + x + self.margin,
					self.y + y + self.margin + self._child.get_height() -
					self._btn.get_height())
	
	def get_width(self):
		if self.get_collapsed() or self._child == None:
			return CollapseWidget.BUTTON_SIZE + 2 * self.margin
		else:
			return 3 * self.margin + self._child.get_width()
	
	def get_height(self):
		if self.get_collapsed() or self._child == None:
			return CollapseWidget.BUTTON_SIZE + 2 * self.margin
		else:
			return 2 * self.margin + self._child.get_height()
		
	def get_scale_width(self):
		if self.get_collapsed():
			return 0
		else:
			return self.get_width() - 3 * self.margin \
					- CollapseWidget.BUTTON_SIZE
	
	def get_scale_height(self):
		if self.get_collapsed():
			return 0
		else:
			return self.get_height() - 2 * self.margin

	def get_scale_left(self):
		return self.margin
	
	def get_scale_bottom(self):
		return self.margin
	
	def set_width(self,val):
		mc = self.evtmngr.MakeChanges()
		
		ContainerWidget.set_width(self,val)

		if self._child != None:
			self._child.set_width(int(val) - 3 * self.margin -
				CollapseWidget.BUTTON_SIZE)
		
	def set_height(self,val):
		mc = self.evtmngr.MakeChanges()
		
		ContainerWidget.set_height(self,val)

		if self._child != None:
			self._child.set_height(int(val) - 2 * self.margin)

	def set_collapsed(self,val):
		"""
		Set the collapsed state.

		@type  val: bool
		@param val: The collapsed state.
		"""
		mc = self.evtmngr.MakeChanges()

		self._btn.set_value( int(not bool( val )) )
		self.collapse( bool( val ) )

	def get_collapsed(self):
		"""
		Get the collapsed state.

		@rtype:  bool
		@return: The collapsed state.
		"""
		return not bool( self._btn.get_value() )

	def clear(self):
		"""
		Clears the child widget.
		"""
		if self._child != None:
			self.remove(self._child)
		self._child = None

	def set_child(self,child):
		"""
		Set the child widget.

		@type  child: Widget or None
		@param child: The new child widget.
		"""
		mc = self.evtmngr.MakeChanges()

		if self._child != None:
			self.remove(self._child)
		
		if child != None:
			self.add(child)
			child.x = 0
			child.y = 0
			child.set_height(int(val) - 2 * self.margin)
			child.set_width( int(val) - 3 * self.margin -
				CollapseWidget.BUTTON_SIZE)
		
		self._child = child
	
	def get_child(self):
		"""
		Get the child widget or None.

		@rtype: Widget or None
		@return: The child widget or None.
		"""
		return self._child

class WidgetFrame(ContainerWidget):
	"""
	Create a WidgetFrame.

	All coordinates of child-widgets in a WidgetFrame are relative to the
	coordinates of the WidgetFrame (be aware of the margin!).

	Futher are childs only drawn if the WidgetFrame is visible.
	"""
	def __init__(
			self,
			childs = [],
			margin = 5,
			bgcolor = GREY,
			x = 0,
			y = 0,
			width   = 100,
			height  = 100,
			parent  = None,
			evtmngr = None):
		"""
		@type  childs: list
		@param childs: list of child-widgets.
		@type  margin: int
		@param margin: The margin around child-widgets.
		@type  bgcolor: ScriptUI.Color.Color
		@param bgcolor: The background-color.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		ContainerWidget.__init__(self,childs,x,y,width,height,parent,evtmngr)

		self.bgcolor = Color(*bgcolor)
		self.margin  = int(margin)
		
	def draw(self,x,y):
		if self.visible:
			cld_x    = self.x + x + self.margin
			cld_y    = self.y + y + self.margin
			oldcolor = getCurrentColor()

			glColor4f( *self.bgcolor )
			glRecti( self.x + x, self.y + y, self.x + x + self._width, self.y + y + self._height )
			glColor4f( *oldcolor )
		
			for child in self:
				child.draw(cld_x,cld_y)

	def get_scale_width(self):
		return self._width - 2 * self.margin
	
	def get_scale_height(self):
		return self._height - 2 * self.margin
	
	def get_scale_left(self):
		return self.margin
	
	def get_scale_bottom(self):
		return self.margin

class Rect(Widget):
	"""
	Draw a simple rectangle to the screen.
	"""
	def __init__(
			self,
			color = GREY,
			x = 0,
			y = 0,
			width   = 100,
			height  = 100,
			parent  = None,
			evtmngr = None):
		"""
		@type  color: ScriptUI.Color.Color
		@param color: The color of the rectangle.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		Widget.__init__(self,x,y,width,height,parent,evtmngr)

		self.color = Color(*color)

	def draw(self,x,y):
		if self.visible:
			oldcolor = getCurrentColor()

			glColor4f( *self.color )
			glRecti( self.x + x, self.y + y, self.x + x + self._width, self.y + y + self._height )
			glColor4f( *oldcolor )

class WidgetList(ContainerWidget):
	"""
	Create a WidgetList.

	All childs of a WidgetList will be placed and resized automatically.
	"""
	def __init__(
			self,
			childs  = [],
			margin  = 5,
			spacing = 3,
			widget_width  = -1,
			widget_height = -1,
			vertical = True,
			reverse  = False,
			x = 0,
			y = 0,
			width   = 100,
			height  = 100,
			parent  = None,
			evtmngr = None):
		"""
		@type  childs: list
		@param childs: list of child-widgets.
		@type  margin: int
		@param margin: The margin around child-widgets.
		@type  spacing: int
		@param spacing: The space between the child-widgets.
		@type  widget_width: int
		@param widget_width: The width of the child-widgets. Will be used wehn vertical == True.
			widget_width == -1 means do not change the width of the child-widgets.
		@type  widget_height: int
		@param widget_height: The height of the child-widgets. Will be used wehn vertical == False.
			widget_height == -1 means do not change the height of the child-widgets.
		@type  vertical: bool
		@param vertical: True to order the child-widget vertical (default), False to order them horizontal.
		@type  reverse: bool
		@param reverse: True to reverse the order of the widgets.
		@type  x: int
		@param x: x position
		@type  y: int
		@param y: y position
		@type  width:  int
		@type  height: int
		@type  parent: Widget
		@param parent: The parent widget.
		@type  evtmngr: ScriptUI.EventManager
		@param evtmngr: The EventManager to use.
		"""
		self.margin        = int(margin)
		self.spacing       = int(spacing)
		self.widget_width  = int(widget_width)
		self.widget_height = int(widget_height)
		self.vertical      = bool(vertical)
		self.reverse       = bool(reverse)
		self._changed      = True
		self._old_x        = int(x)
		self._old_y        = int(y)
		self._old_rev      = bool(reverse)
		self._old_vert     = bool(vertical)
		
		ContainerWidget.__init__(self,childs,x,y,width,height,parent,evtmngr)
	
	def draw(self,x,y):
		if self._old_x != self.x or self._old_y != self.y or \
				self._changed or self._old_rev != self.reverse or \
				self._old_vert != self.vertical:
			self._changed  = False
			self._old_x    = self.x
			self._old_y    = self.y
			self._old_rev  = self.reverse
			self._old_vert = self.vertical
			self.reorder()
		ContainerWidget.draw(self,x,y)
	
	def add(self,child):
		mc = self.evtmngr.MakeChanges()
		
		ContainerWidget.add(self,child)
		self._changed = True
		
	def remove(self,child):
		mc = self.evtmngr.MakeChanges()
		
		ContainerWidget.remove(self,child)
		self._changed = True

	def reorder(self):
		"""
		Reordes the child-widgets.
		"""
		w = self.get_width()
		h = self.get_height()
		
		y = self.margin
		x = self.margin

		if self.vertical:
			w -= 2 * self.margin

			if self.reverse:
				y = self._height - self.margin

			for wdg in self:
				wdg_h = wdg.get_height()
				wdg.x = x
				wdg.y = y

				if self.reverse:
					wdg.y -= wdg_h
					y     -= wdg_h + self.spacing
				else:
					y     += wdg_h + self.spacing
					
				try:
					wdg.set_width( w )
				except AttributeError:
					pass
		else:
			h -= 2 * self.margin

			if self.reverse:
				x = self._width - self.margin

			for wdg in self:
				wdg_w = wdg.get_width()
				wdg.y = y
				wdg.x = x

				if self.reverse:
					wdg.x -= wdg_w
					x     -= wdg_w + self.spacing
				else:
					x     += wdg_w + self.spacing

				try:
					wdg.set_height( h )
				except AttributeError:
					pass

	def set_width(self,val):
		mc = self.evtmngr.MakeChanges()
		
		ContainerWidget.set_width(self,val)
		self._changed = True
		
	def set_height(self,val):
		mc = self.evtmngr.MakeChanges()
		
		ContainerWidget.set_height(self,val)
		self._changed = True
	
	def get_width(self):
		if self.vertical:
			return self._width
		else:
			w = 2 * self.margin

			if not self.is_empty():
				for wdg in self:
					if self.widget_width >= 0:
						try:
							wdg.set_width( self.widget_width )
						except AttributeError:
							pass
						
					w += wdg.get_width() + self.spacing
				w -= self.spacing
			return w
				
	def get_height(self):
		if not self.vertical:
			return self._height
		else:
			h = 2 * self.margin

			if not self.is_empty():
				for wdg in self:
					if self.widget_height >= 0:
						try:
							wdg.set_height( self.widget_height )
						except AttributeError:
							pass
					
					h += wdg.get_height() + self.spacing
				h -= self.spacing
			return h

def split_to_fit(line,draw_width,style):
	"""
	Splits line into parts which are at most draw_width large.

	@type  line: string
	@param line: The line to split.
	@type  draw_width: int
	@param draw_width: The maximal allowed line-width.
	@type  style: string
	@param style: The size of the font: 'large', 'normal', 'small' or 'tiny'.

	@rtype:  list
	@return: The splitted text.
	"""
	if len(line) == 0:
		return ['']

	if draw_width <= 0:
		return [c for c in line]
	
	lines = []
	while len(line) > 0:
		llen = Blender.Draw.GetStringWidth( line, style )

		if llen <= draw_width:
			lines.append( line )
			break
		else:
			prev = 0
			end  = next_word_end( line, 0 )

			while end != -1 and Blender.Draw.GetStringWidth( line[:end], style ) < draw_width:
				prev = end
				end  = next_word_end( line, end )

			if prev == 0:
				end = 1
				while Blender.Draw.GetStringWidth( line[:end], style ) < draw_width:
					end += 1
				prev = end - 1

				if prev == 0: prev = 1
					
			lines.append( line[:prev] )
			line = line[prev:]

	return lines

def next_word_end( text, start ):
	"""
	Finds the end of the next word after position start.
	
	@type  text: string
	@param text: Text to search in.
	@type  start: int
	@param start: The start postion to begin the search.

	@rtype:  int
	@return: the position of the end of the next word. -1 if there are no more words.
	"""
	slen = len(text)

	while start < slen and text[start].isspace():
		start += 1
		
	while start < slen and not text[start].isspace():
		start += 1
	
	if start >= slen:
		return -1
	else:
		return start
