#!BPY
# vim:encoding=ISO-8859-1

"""
Wrapper to blenders script-UI-eventsystem.

@version: 0.6.2
@author:  Mathias Panzenbck
@license: LGPL

@group Event Management: EventManager, MakeChanges

@group Helper Classes: Functor, MultiFunctor, MenueBase

@group Helper Functions: NOOP
"""

__author__  = "Mathias Panzenbck"
__version__ = "0.6.2"
__url__ = ("blender","elysiun","http://www.elysiun.com/forum/viewtopic.php?t=38458&postdays=0&postorder=asc&start=0",)
__license__ = "LGPL"
__bpydoc__ = """\
Description: Wrapper to blenders script-UI-eventsystem
Usage: only used by other scripts
Notes:
"""

# --------------------------------------------------------------------------
# ScriptUI, Copyright (C) 2005  Mathias Panzenbck
# --------------------------------------------------------------------------
# ***** BEGIN LGPL LICENSE BLOCK *****
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# ***** END LGPL LICENCE BLOCK *****
# -------------------------------------------------------------------------- 

import Blender

MAX_KEY_EVENT = 311

def NOOP(*args):
	"""
	Dummy function.

	Use this function whenever a callback is needed but you don't
	want to pass one.

	@type  args: argument-list
	@param args: Will be ignored.
	"""
	pass
	
class Functor:
	"""
	A simple functor class.

	This class is a functor which calls a referenced function
	with a value stored as class-member as first parameter and
	all parameters passed to the L{__call__} method of the functor
	after that.
	"""
	
	def __init__(
			self,
			funct_clbk = NOOP,
			data = None):
		"""
		@type  funct_clbk: function with at least one parameter
		@param funct_clbk: The function which will be called.
		@param       data: First parameter of funct_clbk
		"""
		self.funct_clbk = funct_clbk
		self.data       = data
	
	def __call__(self,*args):
		"""
		@param args: The function arguments.
		@return:     The return value of the callback function.
		"""
		return self.funct_clbk(self.data,*args)

class MultiFunctor:
	"""
	Calls multible functions per call.

	@ivar callbacks: List of either tupels (callback, first_arg) or
		just callbacks. (Can be mixed, too.)
	@type callbacks: list
	"""
	def __init__(
			self,
			callbacks = []):
		"""
		@type  callbacks: list
		@param callbacks: List of either tupels (callback, first_arg) or
			just callbacks. (Can be mixed, too.)
		"""
		self.callbacks = callbacks
	
	def __call__(self,*args):
		"""
		@param args: The function arguments.
		"""
		for clbk in self.callbacks:
			if type(clbk) == tuple:
				clbk[0](clbk[1],*args)
			else:
				clbk(*args)

class MenuBase:
	"""
	Backend-class for any kind of menu.

	@cvar ITEM: Defines a menu-entry as "item".
	@type ITEM: int
	@cvar SEP:  Defines a menu-entry as "seperator".
	@type SEP:  int
	"""
	ITEM = 0
	SEP  = 1

	def __init__(self,name=None,items=[],cancel_clbk=NOOP):
		"""
		Initialise a menu.

		To initialise a menu pass a list of tupels as the items parameter.
		
		For a menu-entry the tuple has to be: C{(MenuBase.ITEM, ("name", callback))}
		The function callback will be called (with no parameter) if the given
		menu-entry was clicked.
		
		For a seperator the tuple has to be: C{(MenuBase.SEP,)}

		@type  name: string or None
		@param name: The name (title) of the menu or None, if no title is wanted.
		@type  items: list
		@param items: The items of the menu.
		@type  cancel_clbk: function with no parameter
		@param cancel_clbk: This function will be called if the selection will be canceld.
		"""
		if name == None:
			self.name = None
		else:
			self.name  = str(name)

		self.items     = []
		self.callbacks = {-1:cancel_clbk}

		for item in items:
			if item[0] == MenuBase.ITEM:
				self.addItem(*item[1])
			else:
				self.addSep()
	
	def addItem(self,name,clbk=NOOP):
		"""
		Add a new menu-entry.

		@type  name: string
		@param name: The name (title) of the new menu-entry.
		@type  clbk: function with no parameters
		@param clbk: The function to call when the menu-entry was clicked.
		"""
		self.callbacks[ len(self.items) ] = clbk
		self.items.append( (MenuBase.ITEM, name) )

	def addSep(self):
		"""
		Add a new seperator.
		"""
		self.items.append( (MenuBase.SEP,) )

	def clear(self):
		"""
		Clears all menu-entries and callbacks.
		"""
		self.items     = []
		self.callbacks = {}

	def get_menu_str(self):
		"""
		Returns a string which you can use in a Blender.Draw.Menu
		or Blender.Draw.PupMenu.

		@rtype:  string
		@return: The menu-string.
		"""
		if self.name == None:
			mnu_str = ""
		else:
			mnu_str = self.name + "%t"
			
		i = 0
		for itm in self.items:
			if itm[0] == MenuBase.ITEM:
				mnu_str += "|" + itm[1] + "%x" + str(i)
			else:
				mnu_str += "|%l"
			i += 1

		return mnu_str

	def __call__(self,ind):
		"""
		Call this to signal that the menu-entry nr. C{ind} was clicked.

		@type  ind: int
		@param ind: The index of the clicked menu-entry (-1 for cancel).
		"""
		self.callbacks[ ind ]()
	
class MakeChanges:
	"""
	This class helps you to invoke a redraw.
	
	Whenever you make changes to a widget in a widget-member-function which would
	change the apperance of it, make a instance of this class.
	
	Alternatively you clould call your L{EventManager}s L{EventManager.begin_change}
	before and L{EventManager.end_change} method after the changes you make.
	These functions are called within the constructor and destructor of the MakeChanges
	class, what provides more safety about that end_change() is called anyway (what
	if there is an exception thrown and your end_change() call is never reached?).

	Example code::
		def member_of_widget_which_makes_changes(self,*args):
			mc = MakeChanges(self.evtmngr)
			# make changes...

	When the mc object is destroyed (at the end of the function), the redraw will be triggered.
	But when you call another function, which makes changes in that way, in this function, the redraw will
	be triggered when the last MakeChanges object is destroyed. So never store such a object as a
	class member. Assure that only one reference to each MakeChanges object exists, so the refcounting
	garbadge collector can call the destructor at the right time.
	"""
	_instances = 0

	def __init__(self,evtmngr):
		"""
		@type  evtmngr: L{EventManager}
		@param evtmngr: Your EventManger.
		"""
		MakeChanges._instances += 1
		self.evtmngr = evtmngr
		self.evtmngr.begin_change()
	
	def __del__(self):
		MakeChanges._instances -= 1
		self.evtmngr.end_change()

import ScriptUI.Widgets
from ScriptUI.Widgets import *

import ScriptUI.Color
from ScriptUI.Color import *

class EventManager:
	"""
	Use this class to manage all the events in your script.
	
	It stores two dictionaries, one	for button-events and one for key-events (just called events here).
	It also manages the redraws	and stores therefor a reference to the root-widget.

	@ivar root: The root-widget. All other widgets must be childs or offspring of it, so they will be drawn.
	"""
	def __init__(self):
		self.root = None
		self.clear()
		self.root   = RootWidget(self)
		self.noroot = RootWidget(self)

	def clear(self):
		"""
		Clear all values (key-events, button-events, draw-callbacks, ...).
		"""
		self._event     = {}
		self._bevent    = {}
		self._evt_maxid = MAX_KEY_EVENT
		self._draw      = []
		self._todraw    = 0
		self._pending   = False

	def __del__(self):
		self.clear()

	def MakeChanges(self):
		"""
		Create a L{MakeChanges} class without passing the EventManager as parameter.

		@rtype: MakeChanges
		"""
		# no need to create a MakeChanges object if more then 0 instances allready exists
		if MakeChanges._instances > 0:
			return None
		else:
			return MakeChanges(self)

	def run(self):
		"""
		This function starts the GUI.
		"""
		Blender.Draw.Register(self.draw, self.event, self.button_event)
	
	def exit(self):
		"""
		This function exits the GUI.
		
		Allways use this function, never use Blender.Draw.Exit(), because if you would do that,
		no widget would be destroyed (memory leak!).
		"""
		self.root.unref()
		self.noroot.unref()
		self.root   = None
		self.noroot = None
		self.clear()
		Blender.Draw.Exit()
	
	def register_draw(self,draw):
		"""
		Registers a function so it will be caled at every redraw.
		
		(Don't know if this functionallity makes much sense.)

		@type  draw: function with no parameters
		@param draw: A callback function which will be called at every redraw.
		"""
		self._draw.append(draw)
	
	def unregister_draw(self,draw):
		"""
		Deletes the given function from the list of functions which will be called at every redraw.

		@type  draw: function with no parameters
		@param draw: A callback function.
		"""
		self._draw.remove(draw)
	
	def register_button_event(self,btn_event_clbk):
		"""
		Registes a new event callback, used for button-events, and returns the event-number.

		@type  btn_event_clbk: function with one parameter, the event-number (int)
		@param btn_event_clbk: The callback function.

		@rtype: int
		@return: The event-number.
		"""
		self._evt_maxid += 1
		self._bevent[self._evt_maxid] = btn_event_clbk
		return self._evt_maxid

	def unregister_button_event(self,event):
		"""
		Deletes the given button-event.

		@type  event: int
		@param event: The event-number to unregister.
		"""
		del self._bevent[event]
	
	def register_event(self,event,event_clbk):
		"""
		Registers a callback function to a key-event.

		@type       event: int
		@param      event: The key-event (keycode).
		@type  event_clbk: function with two parameters: event (int), modifier (int)
		@param event_clbk: The function to call when given key-event occures.

		@raise KeyError: When event > L{MAX_KEY_EVENT}.
		"""
		if event > MAX_KEY_EVENT:
			raise KeyError, "illegal eventnumber"
		
		try:
			evtlist = self._event[event]
			evtlist.append( event_clbk )
		except KeyError:
			self._event[event] = [ event_clbk ]
	
	def unregister_event(self,event,event_clbk):
		"""
		Deletes given callback function from given key-event.
		
		@type       event: int
		@param      event: The key-event (keycode).
		@type  event_clbk: function with two parameters: event (int), modifier (int)
		@param event_clbk: The function to call when given key-event occures.
		"""
		evtlist = self._event[event]
		evtlist.remove(event_clbk)

		if evtlist == []:
			del self._event[event]
	
	def clear_event(self,event):
		"""
		Deletes all callbacks from the given key-event.
		
		@type       event: int
		@param      event: The key-event (keycode).
		"""
		del self._event[event]
	
	def event(self,evt,modi):
		"""
		The key-event callback function passed to blenders Blender.Draw.Register() function.

		B{Do not call this function unless you know exactly what you are doing.}
		"""
		try:
			for event_clbk in self._event[evt]:
				event_clbk(evt,modi)
		except KeyError:
			pass
	
	def button_event(self,evt):
		"""
		The button-event callback function passed to blenders Blender.Draw.Register() function.

		B{Do not call this function unless you know exactly what you are doing.}
		"""
		try:
			self._bevent[evt](evt)
		except KeyError:
			print "*** KeyError:", evt
	
	def draw(self):
		"""
		The draw callback function passed to blenders Blender.Draw.Register() function.

		B{Do not call this function unless you know exactly what you are doing.}
		"""
		self._pending = True

		for draw_clbk in self._draw:
			draw_clbk()

		self.root.draw(0,0)
		self._pending = False

	def get_pending(self):
		"""
		Returns True if there is a pending redraw.

		@rtype: bool
		@return: True if there is a pending redraw.
		"""
		return self._pending

	def begin_change(self):
		"""
		Call this function whenever you begin to make changes, which will change the
		apperance of your derived widget.

		Better use teh L{MakeChanges} class instead.
		"""
		self._todraw += 1

	def end_change(self):
		"""
		Call this function whenever you end to make changes, which will change the
		apperance of your derived widget.

		Better use the L{MakeChanges} class instead.
		"""
		self._todraw -= 1

		if self._todraw <= 0:
			self._todraw = 0
			
# XXX: redraws are buffered anyway. so, is this wise?
# however, this prevents endless recursions, when you
# call begin- and end_cange() from within a draw()
# method.
			if not self._pending:
				Blender.Draw.Redraw()
	
	def Spacer(
			self,
			x = 0,
			y = 0,
			width  = 100,
			height = 100,
			parent = None):
		"""
		Create a empty widget (L{ScriptUI.Widgets.Widget}) without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.Widget
		"""
		return Widget(x,y,width,height,parent,self)
	
	def PushButton(
			self,
			name,
			x = 0,
			y = 0,
			width   = 100,
			height  = 100,
			bevent  = NOOP,
			tooltip = "",
			parent  = None):
		"""
		Create a L{ScriptUI.Widgets.PushButton} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.PushButton
		"""
		parent = parent or self.root
		return PushButton(name,x,y,width,height,bevent,tooltip,parent,self)

	def Menu(self,name,x=0,y=0,width=100,height=100,
			items=[],initvalue=0,tooltip="",parent=None):
		"""
		Create a L{ScriptUI.Widgets.Menu} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.Menu
		"""
		parent = parent or self.root
		return Menu(name,x,y,width,height,items,initvalue,tooltip,parent,self)

	def Number(self,name,x=0,y=0,width=100,height=100,
			bevent=NOOP,initvalue=1.,min=0.,max=1.,tooltip="",parent=None):
		"""
		Create a L{ScriptUI.Widgets.Number} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.Number
		"""
		parent = parent or self.root
		return Number(name,x,y,width,height,bevent,initvalue,min,max,tooltip,parent,self)

	def Slider(self,name,x=0,y=0,width=100,height=100,
			bevent=NOOP,initvalue=1.,min=0.,max=1.,realtime=0,tooltip="",parent=None):
		"""
		Create a L{ScriptUI.Widgets.Slider} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.Slider
		"""
		parent = parent or self.root
		return Slider(name,x,y,width,height,bevent,initvalue,min,max,realtime,tooltip,parent,self)

	def String(self,name,x=0,y=0,width=100,height=100,
			bevent=NOOP,initvalue="",length=512,tooltip="",parent=None):
		"""
		Create a L{ScriptUI.Widgets.String} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.String
		"""
		parent = parent or self.root
		return String(name,x,y,width,height,bevent,initvalue,length,tooltip,parent,self)

	def Toggle(self,name,x=0,y=0,width=100,height=100,bevent=NOOP,initvalue=0,tooltip="",parent=None):
		"""
		Create a L{ScriptUI.Widgets.Toggle} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.Toggle
		"""
		parent = parent or self.root
		return Toggle(name,x,y,width,height,bevent,initvalue,tooltip,parent,self)

	def SimpleText(
			self,
			text  = "",
			style = "normal",
			line_height = 15,
			bottom_up   = False,
			color = WHITE,
			x = 0,
			y = 0,
			parent = None):
		"""
		Create a L{ScriptUI.Widgets.SimpleText} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.SimpleText
		"""
		parent = parent or self.root
		return SimpleText(text,style,line_height,bottom_up,color,x,y,parent,self)
	
	def ScrollText(
			self,
			text  = "",
			style = "normal",
			line_height = 15,
			margin  = 5,
			color   = BLACK,
			bgcolor = GREY,
			x = 0,
			y = 0,
			width  = 100,
			height = 36,
			parent = None):
		"""
		Create a L{ScriptUI.Widgets.ScrollText} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.ScrollText
		"""
		parent = parent or self.root
		return ScrollText(text,style,line_height,margin,color,bgcolor,x,y,width,height,parent,self)

	def WidgetFrame(
			self,
			childs = [],
			margin = 5,
			bgcolor = GREY,
			x = 0,
			y = 0,
			width  = 100,
			height = 100,
			parent = None):
		"""
		Create a L{ScriptUI.Widgets.WidgetFrame} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.WidgetFrame
		"""
		parent = parent or self.root
		return WidgetFrame(childs,margin,bgcolor,x,y,width,height,parent,self)

	def Rect(
			self,
			color = GREY,
			x = 0,
			y = 0,
			width  = 100,
			height = 100,
			parent = None):
		"""
		Create a L{ScriptUI.Widgets.Rect} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.Rect
		"""
		parent = parent or self.root
		return Rect(color,x,y,width,height,parent,self)

	def WidgetList(
			self,
			childs  = [],
			margin  = 5,
			spacing = 3,
			widget_width  = -1,
			widget_height = -1,
			vertical = True,
			reverse  = False,
			x = 0,
			y = 0,
			width  = 100,
			height = 100,
			parent = None):
		"""
		Create a L{ScriptUI.Widgets.WidgetList} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.WidgetList
		"""
		parent = parent or self.root
		return WidgetList(childs,margin,spacing,
				widget_width,widget_height,
				vertical,reverse,
				x,y,width,height,
				parent,self)
	
	def FileField(
			self,
			caption = "File: ",
			browse_caption = "Open File",
			path    = "",
			x = 0,
			y = 0,
			width    = 200,
			height   =  20,
			action   = NOOP,
			selector = FileField.FILE,
			tooltip  = "",
			parent   = None):
		"""
		Create a L{ScriptUI.Widgets.FileField} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.FileField
		"""
		parent = parent or self.root
		return FileField(caption,browse_caption,path,x,y,width,height,
				action,selector,tooltip,parent,self)
	
	def Image(self,image,
			x = 0, y = 0,
			zoomx = 1.0, zoomy = 1.0,
			clipx = 0, clipy = 0,
			clipw = 100, cliph = 100,
			parent=None):
		"""
		Create a L{ScriptUI.Widgets.Image} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.Image
		"""
		parent = parent or self.root
		return Image(image,x,y,
				zoomx,zoomy,
				clipx,clipy,clipw,cliph,
				parent,self)
	
	def PixelField(self,
			x=0,y=0,
			width=100,height=100,
			parent=None):
		"""
		Create a L{ScriptUI.Widgets.PixelField} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.PixelField
		"""
		parent = parent or self.root
		return PixelField(x,y,width,height,parent,self)

	def RenderField(self,render_clbk,
			x=0,y=0,
			width=100,height=100,
			parent=None):
		"""
		Create a L{ScriptUI.Widgets.RenderField} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.RenderField
		"""
		parent = parent or self.root
		return RenderField(render_clbk,
				x,y,width,height,
				parent,self)

	def CollapseWidget(self,
		x = 0,
		y = 0,
		width    = 100,
		height   = 100,
		margin   = 5,
		child    = None,
		collapse = NOOP,
		parent   = None):
		"""
		Create a L{ScriptUI.Widgets.CollapseWidget} without passing the EventManager as parameter.
		
		@rtype: ScriptUI.Widgets.CollapseWidget
		"""
		parent = parent or self.root
		return CollapseWidget(x,y,width,height,
				margin,child,collapse,
				parent,self)
